from hyponic.optimizers.base_optimizer import BaseOptimizer

import numpy as np
import numexpr as ne


class GA(BaseOptimizer):
    """
    Genetic Algorithm(GA)

    Example
    ~~~~~~~
        >>> from hyponic.optimizers.genetic_based.GA import GA
        >>> import numpy as np
        >>>
        >>> def sphere(x):
        >>>     return np.sum(x ** 2)
        >>>
        >>> problem_dict = {
        >>>     'fit_func': sphere,
        >>>     'lb': [-5.12, -5, -14, -6, -0.9],
        >>>     'ub': [5.12, 5, 14, 6, 0.9],
        >>>     'minmax': 'min'
        >>> }
        >>>
        >>> ga = GA(epoch=40, population_size=100, verbose=True, early_stopping=4)
        >>> ga.solve(problem_dict)
        >>> print(ga.get_best_score())
        >>> print(ga.get_best_solution())
    """

    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None,
                 verbose: bool = False, mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None,
                 **kwargs):
        """
                :param epoch: number of iterations
                :param population_size: number of individuals in the population
                :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
                :param verbose: whether to print the progress, default is False
                :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
                :param n_workers: number of workers to use in multithreading mode
                :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
                stopping is not used
        """
        super().__init__(**kwargs)
        self.epoch = epoch
        self.population_size = population_size
        self.minmax = minmax
        self.verbose = verbose
        self.mode = mode
        self.n_workers = n_workers
        self.early_stopping = early_stopping

        self.population = None
        self.scores = None
        self.best_score = None
        self.best_solution = None

    def initialize(self, problem_dict):
        super().initialize(problem_dict)

        self.population = np.random.uniform(low=self.lb, high=self.ub, size=(self.population_size, self.dimensions))
        self.scores = np.array([self.function(self.population[i]) for i in range(self.population_size)])

        best_idx = self._argminmax()(self.scores)
        self.best_score = self.scores[best_idx]
        self.best_solution = self.population[best_idx]

    def evolve(self, epoch):
        next_population = np.zeros_like(self.population)
        next_scores = np.zeros(self.population_size)

        # Elitism: keep the best solution from the previous generation
        best_idx = self._argminmax()(self.scores)
        next_population[0] = self.population[best_idx]
        next_scores[0] = self.scores[best_idx]

        # Roulette Wheel Selection
        fitness = self.scores - np.min(self.scores) if self.minmax == 'min' else np.max(self.scores) - self.scores
        total_fitness = np.sum(fitness)
        if total_fitness == 0:
            probs = np.ones(self.population_size) / self.population_size
        else:
            probs = fitness / total_fitness
        cum_probs = np.cumsum(probs)

        # Crossover and mutation
        for i in range(1, self.population_size):
            # Select two parents using roulette wheel selection
            parent1_idx = np.searchsorted(cum_probs, np.random.rand())
            parent2_idx = np.searchsorted(cum_probs, np.random.rand())

            # Single point crossover
            crossover_point = np.random.randint(1, self.dimensions)

            next_population[i, :crossover_point] = self.population[parent1_idx, :crossover_point]
            next_population[i, crossover_point:] = self.population[parent2_idx, crossover_point:]

            # Mutation
            mutation_strength = ne.evaluate("0.5 * (ub - lb)", local_dict={"ub": self.ub, "lb": self.lb})

            next_population[i] += np.random.normal(0, mutation_strength, size=self.dimensions)
            next_population[i] = np.clip(next_population[i], self.lb, self.ub)

        # evaluate the new population
        next_scores = np.array([self.function(next_population[i]) for i in range(self.population_size)])

        # update the best solution and score
        best_idx = self._argminmax()(next_scores)
        if self._minmax()(next_scores) < self._minmax()(self.scores):
            self.best_solution = next_population[best_idx]
            self.best_score = next_scores[best_idx]

        # replace the old population with the new one
        self.population = next_population
        self.scores = next_scores

    def get_best_score(self):
        return self.best_score

    def get_best_solution(self):
        return self.best_solution

    def get_current_best_score(self):
        return self._minmax()(self.scores)

    def get_current_best_solution(self):
        return self.population[self._argminmax()(self.scores)]



"""
This module contains metrics for evaluating classification models.
"""
from hyponic.metrics.decorators import add_metric_info, maximize_metric, minimize_metric

import numpy as np
import numba as nb


@maximize_metric
@add_metric_info("Accuracy is the fraction of predictions model got right.")
@nb.njit
def accuracy(y_true: np.array, y_pred: np.array) -> np.ndarray:
    return np.mean(np.equal(y_true, y_pred))


@maximize_metric
@add_metric_info("Precision is the fraction of positive predictions that are correct.")
@nb.njit
def precision(y_true: np.array, y_pred: np.array) -> np.ndarray | float:
    tp = np.sum(np.logical_and(y_true == 1, y_pred == 1))
    fp = np.sum(np.logical_and(y_true == 0, y_pred == 1))
    if tp + fp == 0:
        return 0.0
    return tp / (tp + fp)


@maximize_metric
@add_metric_info("Recall is the fraction of positive predictions that are correct")
@nb.njit
def recall(y_true: np.array, y_pred: np.array) -> np.ndarray | float:
    tp = np.sum(np.logical_and(y_true == 1, y_pred == 1))
    fn = np.sum(np.logical_and(y_true == 1, y_pred == 0))
    if tp + fn == 0:
        return 0.0
    return tp / (tp + fn)


@maximize_metric
@add_metric_info("F1 score is the harmonic mean of precision and recall.")
@nb.njit
def f1_score(y_true: np.array, y_pred: np.array) -> np.ndarray | float:
    p = precision(y_true, y_pred)
    r = recall(y_true, y_pred)
    if p + r == 0:
        return 0.0
    return 2 * p * r / (p + r)


@maximize_metric
@add_metric_info("F-beta score is the weighted harmonic mean of precision and recall.")
@nb.njit
def fbeta(y_true: np.array, y_pred: np.array, beta: float = 1.0) -> np.ndarray | float:
    p = precision(y_true, y_pred)
    r = recall(y_true, y_pred)
    if p + r == 0:
        return 0.0
    return (1 + beta ** 2) * p * r / (beta ** 2 * p + r)


@maximize_metric
@add_metric_info(
    "Matthews correlation coefficient is a correlation coefficient between the observed and predicted binary classifications.")
@nb.njit
def confusion_matrix(y_true: np.array, y_pred: np.array) -> np.ndarray:
    tp = np.sum(np.logical_and(y_true == 1, y_pred == 1))
    fp = np.sum(np.logical_and(y_true == 0, y_pred == 1))
    fn = np.sum(np.logical_and(y_true == 1, y_pred == 0))
    tn = np.sum(np.logical_and(y_true == 0, y_pred == 0))
    return np.array([[tp, fp], [fn, tn]])


@minimize_metric
@add_metric_info(
    "Log loss is the negative log-likelihood of the true labels given a probabilistic classifier’s predictions.")
@nb.njit
def log_loss(y_true: np.array, y_pred: np.array) -> np.ndarray:
    return np.mean(np.log(1 + np.exp(-y_true * y_pred)))


@minimize_metric
@add_metric_info(
    "Binary crossentropy is the negative log-likelihood of the true labels given a probabilistic classifier’s predictions.")
# @nb.njit
def binary_crossentropy(y_true: np.array, y_pred: np.array) -> np.ndarray:
    # TODO: fix that logarithm can be negative or zero
    return -np.mean(y_true * np.log(y_pred) + (1 - y_true) * np.log(1 - y_pred))


@minimize_metric
@add_metric_info(
    "Categorical crossentropy is the negative log-likelihood of the true labels given a probabilistic classifier’s predictions.")
@nb.njit
def categorical_crossentropy(y_true: np.array, y_pred: np.array) -> np.ndarray:
    return -np.mean(np.sum(y_true * np.log(y_pred), axis=-1))



"""
This module contains the Space class, which is used to define the search space.
"""
import numpy as np

from typing import Callable, Any


class Space:
    """
    A class that represents the search space of a hyperparameter optimization problem.
    """
    def __init__(self, in_dict: dict[str, Any]):
        self.__dict = dict
        self.dimensions = {}
        self.dimensions_names = []

        for k, v in in_dict.items():
            # Converting range to list
            if isinstance(v, range):
                v = list(v)

            if isinstance(v, Discrete):
                self.dimensions[k] = v
            elif isinstance(v, Continuous):
                self.dimensions[k] = v
            elif isinstance(v, tuple):
                if len(v) != 2:
                    raise ValueError(f"Value for key {k} is not valid")

                self.dimensions[k] = Continuous(*v, name=k)
            elif isinstance(v, list):
                self.dimensions[k] = Discrete(v, name=k)
            else:
                raise ValueError(f"Value for key {k} is not valid")

            self.dimensions_names.append(k)

    def get_continuous_mappings(
            self, scales: dict | int | float = None, origins: dict | int | float = None
    ) -> dict[str, (Callable, (float, float))]:
        """
        Returns a function that maps a discrete value to a continuous value.
        """

        if scales is None:
            scales = {}
        elif isinstance(scales, (int, float)):
            scales = {key: scales for key in self.dimensions}

        if origins is None:
            origins = {}
        elif isinstance(origins, (int, float)):
            origins = {key: origins for key in self.dimensions}

        mappings = {}
        for key in self.dimensions:
            mappings[key] = self.dimensions[key].get_continuous_mapping(
                scales.get(key, 1),
                origins.get(key, None)
            )

        return mappings

    def map_to_original_space(self, values: list[float]) -> dict[str, Any]:
        """
        Maps a list of values from the continuous space to the original space.
        """
        mappings = self.get_continuous_mappings()
        return {
            name: mappings[name][0](value) for name, value in zip(self.dimensions_names, values)
        }

    def __str__(self):
        if len(self.dimensions) == 0:
            return "Hyperparameter Search Space is empty."

        # Pretty table of dimensions
        dim_names = list(self.dimensions.keys())
        offset = max([len(k) for k in dim_names])

        out_strs = [f"Hyperparameter Search Space with {len(dim_names)} dimensions:"]
        out_strs += [
            f"\t{k}: {'-'*(offset - len(k))} {self.dimensions[k]}" for k in dim_names
        ]

        return "\n".join(out_strs)


class Dimension:
    def __init__(self, lbound, ubound, name=None):
        self.name = name

        self._lbound = lbound
        self._ubound = ubound

    def get_continuous_mapping(self, scale=1, origin=None) -> (Callable, (float, float)):
        """
        Returns a function that maps set of values to a continuous value
        """
        assert scale != 0, "Scale cannot be 0."

        if origin is None:
            origin = self._lbound

        low = origin
        high = origin + (self._ubound - self._lbound) * scale

        def mapping_func(x):
            x = np.clip(x, low, high)
            x = self._lbound + (x - origin) / scale
            return self.get_value(x)

        return mapping_func, (low, high)

    def get_value(self, x):
        raise NotImplementedError

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.__str__()


class Continuous(Dimension):
    def __init__(self, low, high, name=None):
        super().__init__(low, high, name)

    def get_value(self, x):
        # Note that x is already in the correct range. No need to clip.
        return x

    def __str__(self):
        return super().__str__() + f"({self._lbound}, {self._ubound})"

    def __repr__(self):
        return self.__str__()


class Discrete(Dimension):
    def __init__(self, values, name=None):
        super().__init__(0, len(values), name)
        self.values = np.array(values)

    def get_value(self, x):
        x = np.clip(x, 0, len(self.values) - 1)
        return self.values[int(x)]

    def __str__(self):
        return super().__str__() + f"{self.values}"

    def __repr__(self):
        return self.__str__()



from hyponic.optimizers.base_optimizer import BaseOptimizer

import numpy as np
import numexpr as ne


class ABC(BaseOptimizer):
    """
    Artificial Bee Colony (ABC) algorithm

    Hyperparameters:
        + limits(int), default=25: the number of trials before abandoning food source

    Example
    ~~~~~~~
    >>> from hyponic.optimizers.swarm_based.ABC import ABC
    >>> import numpy as np
    >>>
    >>> def sphere(x):
    >>>     return np.sum(x ** 2)
    >>>
    >>> problem_dict = {
    >>>     'fit_func': sphere,
    >>>     'lb': [-5.12, -5, -14, -6, -0.9],
    >>>     'ub': [5.12, 5, 14, 6, 0.9],
    >>>     'minmax': 'min'
    >>> }
    >>>
    >>> abc = ABC(epoch=40, population_size=100, verbose=True, early_stopping=4)
    >>> abc.solve(problem_dict)
    >>> print(abc.get_best_score())
    >>> print(abc.get_best_solution())
    """

    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None, limits=25,
                 verbose: bool = False, mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None,
                 **kwargs):
        """
        :param epoch: number of iterations
        :param population_size: number of individuals in the population
        :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
        :param verbose: whether to print the progress, default is False
        :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
        :param n_workers: number of workers to use in multithreading mode
        :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
        stopping is not used
        :param limits: the number of trials before abandoning food source
        """
        super().__init__(**kwargs)
        self.epoch = epoch
        self.population_size = population_size
        self.minmax = minmax
        self.verbose = verbose
        self.mode = mode
        self.n_workers = n_workers
        self.early_stopping = early_stopping

        self.limits = limits
        self.fitness = None
        self.g_best = None
        self.g_best_coords = None
        self.trials = None

    def _before_initialization(self):
        super()._before_initialization()
        if not isinstance(self.limits, int) or self.limits < 1:
            raise ValueError("limits should be a positive integer")

    def initialize(self, problem_dict):
        super().initialize(problem_dict)
        self.g_best = np.inf if self._minmax() == min else -np.inf

        self.fitness = np.array([self.function(self.coords[i]) for i in range(self.population_size)])

        self.trials = np.zeros(self.population_size)

    def _coordinate_update_phase(self, i, k):
        phi = np.random.uniform(-1, 1, self.dimensions)
        new_coords = ne.evaluate("coords + phi * (coords - new_coords)", local_dict={'coords': self.coords[i],
                                                                                     'phi': phi,
                                                                                     'new_coords': self.coords[k]})
        new_coords = np.clip(new_coords, self.lb, self.ub)
        new_fitness = self.function(new_coords)
        if self._minmax()(np.array([self.fitness[i], new_fitness])) != self.fitness[i]:
            self.coords[i] = new_coords
            self.fitness[i] = new_fitness
            self.trials[i] = 0
        else:
            self.trials[i] += 1

    def _employed_bees_phase(self):
        for i in range(self.population_size):
            k = np.random.choice([j for j in range(self.population_size) if j != i])
            self._coordinate_update_phase(i, k)

    def _onlooker_bees_phase(self):
        if np.all(self.fitness == 0):
            probabilities = np.ones(self.population_size) / self.population_size
        else:
            probabilities = self.fitness / np.sum(self.fitness)
        for i in range(self.population_size):
            k = np.random.choice([j for j in range(self.population_size)], p=probabilities)
            self._coordinate_update_phase(i, k)

    def _scout_bees_phase(self):
        for i in range(self.population_size):
            if self.trials[i] > self.limits:
                self.coords[i] = np.random.uniform(self.lb, self.ub, self.dimensions)
                self.fitness[i] = self.function(self.coords[i])
                self.trials[i] = 0

    def evolve(self, epoch):
        self._employed_bees_phase()
        self._onlooker_bees_phase()
        self._scout_bees_phase()

        best_index = self._argminmax()(self.fitness)
        self.g_best = self.fitness[best_index]
        self.g_best_coords = self.coords[best_index]

    def get_best_score(self):
        return self.g_best

    def get_best_solution(self):
        return self.g_best_coords

    def get_current_best_score(self):
        return self._minmax()(self.fitness)

    def get_current_best_solution(self):
        best_index = self._argminmax()(self.fitness)
        return self.coords[best_index]



"""
This module contains decorators for metrics.

The following decorators are available:
- add_metric_to_dict -- metric is added to the dictionary of metrics (called automatically)
- add_metric_aliases -- metric can be called by any of the string aliases
- minimize_metric    -- metric is subject to minimization
- maximize_metric    -- metric is subject to maximization

This module also contains a dictionary of metrics to which metrics are added
when they are decorated. This dictionary is used by the HypONIC class to
retrieve the metrics passed to it as the string literals. This is helpful,
because one metric can have multiple aliases, and the user can pass any of
them to the HypONIC class as a string.
"""
import warnings

from collections import defaultdict
from typing import Callable

"""
A dictionary of aliased metrics to which metrics are added when they are decorated.
"""
METRICS_DICT = defaultdict()


def add_metric_to_dict(metric: Callable) -> Callable:
    """
    A decorator that adds the metric to the dictionary of metrics.
    Called automatically by the minimize_metric and maximize_metric decorators.
    :param metric: the metric that should be added to the dictionary
    """
    METRICS_DICT[metric.__name__] = metric

    if not hasattr(metric, "minmax"):
        warnings.warn(f"Metric {metric.__name__} has no minmax attribute. Using 'min' as default.")
        metric.minmax = "min"

    return metric


def add_metric_aliases(*aliases) -> Callable:
    """
    A decorator that adds aliases to the metric.
    :param aliases: a list of aliases for the metric
    :return: decorated metric
    """

    def decorator(metric: Callable) -> Callable:
        for alias in aliases:
            METRICS_DICT[alias] = metric

        return metric

    return decorator


def add_metric_info(info: str) -> Callable:
    """
    A decorator that adds information about the metric.
    :param info: information about the metric
    :return: decorated metric
    """

    def decorator(metric: Callable) -> Callable:
        metric.info = PrintableProperty(info)
        return metric

    return decorator


def maximize_metric(metric) -> Callable:
    """
    A decorator for metrics that should be maximized.
    :param metric: the metric that should be maximized
    """
    metric.minmax = "max"
    return add_metric_to_dict(metric)


def minimize_metric(metric) -> Callable:
    """
    A decorator for metrics that should be minimized.
    :param metric: the metric that should be minimized
    """
    metric.minmax = "min"
    return add_metric_to_dict(metric)


class PrintableProperty:
    """
    A class for the properties that should either be printed or returned as a string.

    Use case:
    mse.info()  # will print the information about the metric
    mse.info    # will return the information about the metric as a string
    """

    def __init__(self, text):
        self.__text = text

    def __str__(self):
        return self.__text

    def __repr__(self):
        return self.__text

    def __call__(self, *args, **kwargs):
        print(self.__text)
        return self.__text



from hyponic.optimizers.base_optimizer import BaseOptimizer

import numpy as np
import numexpr as ne


class GWO(BaseOptimizer):
    """
    Grey Wolf Optimization (GWO) algorithm

    Example
    ~~~~~~~
    >>> from hyponic.optimizers.swarm_based.GWO import GWO
    >>> import numpy as np
    >>>
    >>> def sphere(x):
    >>>     return np.sum(x ** 2)
    >>>
    >>> problem_dict = {
    >>>     'fit_func': sphere,
    >>>     'lb': [-5.12, -5, -14, -6, -0.9],
    >>>     'ub': [5.12, 5, 14, 6, 0.9],
    >>>     'minmax': 'min'
    >>> }
    >>>
    >>> gwo = GWO(epoch=40, population_size=100, verbose=True, early_stopping=4)
    >>> gwo.solve(problem_dict)
    >>> print(gwo.get_best_score())
    >>> print(gwo.get_best_solution())
    """

    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None, verbose: bool = False,
                 mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None, **kwargs):
        """
        :param epoch: number of iterations
        :param population_size: number of individuals in the population
        :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
        :param verbose: whether to print the progress, default is False
        :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
        :param n_workers: number of workers to use in multithreading mode
        :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
        stopping is not used
        """
        super().__init__(**kwargs)
        self.epoch = epoch
        self.population_size = population_size
        self.minmax = minmax
        self.verbose = verbose
        self.mode = mode
        self.n_workers = n_workers
        self.early_stopping = early_stopping

        self.fitness = None
        self.g_best = None
        self.g_best_coords = None

    def initialize(self, problem_dict):
        super().initialize(problem_dict)

        self.g_best = np.inf if self.minmax == 'min' else -np.inf
        self.g_best_coords = np.zeros(self.dimensions)
        self.fitness = np.array([self.function(x) for x in self.coords], dtype=np.float64)

    def evolve(self, current_epoch):
        a = 2 - current_epoch * (2 / self.epoch)

        for i in range(self.population_size):
            r1 = np.random.rand()
            r2 = np.random.rand()
            A = 2 * a * r1 - a
            C = 2 * r2

            D = np.abs(C * self.coords[np.random.randint(0, self.population_size)] - self.coords[np.random.randint(0, self.population_size)])
            coords_new = self.coords[i] + A * D
            fitness_new = self.function(coords_new)
            if self._minmax()([fitness_new, self.fitness[i]]) == fitness_new:
                self.coords[i] = coords_new
                self.fitness[i] = fitness_new

            if self._minmax()([fitness_new, self.g_best]) == fitness_new:
                self.g_best = fitness_new
                self.g_best_coords = coords_new

    def get_best_score(self):
        return self.g_best

    def get_best_solution(self):
        return self.g_best_coords

    def get_current_best_score(self):
        return self._minmax()(self.fitness)

    def get_current_best_solution(self):
        return self.coords[self._argminmax()(self.fitness)]




from abc import ABC, abstractmethod
import numpy as np
from concurrent.futures import ThreadPoolExecutor, as_completed
from hyponic.utils.history import History
import time


class BaseOptimizer(ABC):
    """
    Base class for all optimizers. All optimizers should inherit from this class
    """

    def __init__(self, **kwargs):
        self.epoch = None
        self.population_size = None

        self.function = None
        self.lb = None
        self.ub = None
        self.minmax = None

        self.intervals = None
        self.dimensions = None
        self.verbose = None
        self.mode = None
        self.n_workers = None

        self.coords = None
        self.early_stopping = None

    def _before_initialization(self):
        """
        This method checks if the problem definition is correct
        """
        if not isinstance(self.epoch, int) or self.epoch < 1:
            raise ValueError("epoch should be a positive integer")

        if not isinstance(self.population_size, int) or self.population_size < 1:
            raise ValueError("population_size should be a positive integer")

        if self.mode not in ['single', 'multithread']:
            raise ValueError("mode should be either 'single' or 'multithread'")

        if self.n_workers < 1:  # TODO: n_workers can be -1, which means use all available cores
            raise ValueError("n_workers should be a positive integer")
        if self.early_stopping is not None:
            if not isinstance(self.early_stopping, int) or self.early_stopping < 1:
                raise ValueError("early_stopping should be a positive integer or None")

    def _check_initialization(self):
        """
        This method checks if the problem definition in initialization function is correct
        """
        if self.lb is None or self.ub is None:
            raise ValueError("lb and ub should be provided")

        if not isinstance(self.lb, np.ndarray) or not isinstance(self.ub, np.ndarray):
            raise ValueError("lb and ub should be numpy arrays")

        if self.lb.shape != self.ub.shape:
            raise ValueError("lb and ub should have the same shape")

        if np.any(self.lb > self.ub):
            raise ValueError("lb should be less than ub")

        if self.minmax not in ['min', 'max']:
            raise ValueError("minmax should be either 'min' or 'max'")

        if self.function is None:
            raise ValueError("function should be provided")

        if not callable(self.function):
            raise ValueError("function should be callable")

    def initialize(self, problem_dict):
        """
        Initialize the optimizer with the problem dictionary

        :param problem_dict: dictionary containing the problem definition
        """
        # Unpack the problem dictionary
        self.minmax = problem_dict['minmax'] if self.minmax is None else self.minmax
        self.function = problem_dict["fit_func"]
        self.lb = np.array(problem_dict["lb"])
        self.ub = np.array(problem_dict["ub"])
        self._check_initialization()

        self.intervals = self.ub - self.lb
        self.dimensions = len(self.lb)

        # TODO: add flag that tracks history or not
        self.history = History(optimizer=self, epoch=self.epoch, population_size=self.population_size)

        # Create the population
        self.coords = self._create_population()

    def _create_individual(self):
        """
        Create an individual

        :return: individual
        """
        return np.random.uniform(self.lb, self.ub, self.dimensions)

    def _create_population(self):
        """
        Create a population of size self.population_size
        This method can be parallelized using multithreading or multiprocessing(but multiprocessing doesn't work now)

        :return: population
        """
        coords = np.zeros((self.population_size, self.dimensions))
        if self.mode == 'multithread':
            with ThreadPoolExecutor(self.n_workers) as executor:
                list_executor = [executor.submit(self._create_individual) for _ in range(self.population_size)]
                for f in as_completed(list_executor):
                    coords[list_executor.index(f)] = f.result()
        else:
            coords = np.array([self._create_individual() for _ in range(self.population_size)])
        return coords

    @abstractmethod
    def evolve(self, current_epoch):
        """
        Evolve the population for one epoch

        :param current_epoch: current epoch number
        """
        pass

    @abstractmethod
    def get_best_score(self):
        """
        Get the best score of the current population

        :return: best score of the fitness function
        """
        pass

    @abstractmethod
    def get_best_solution(self):
        """
        Get the best solution of the current population

        :return: coordinates of the best solution
        """
        pass

    @abstractmethod
    def get_current_best_score(self):
        """
        Get the best score of the current population

        :return: current best score of the fitness function
        """
        pass

    @abstractmethod
    def get_current_best_solution(self):
        """
        Get the best solution of the current population

        :return: current coordinates of the best solution
        """
        pass

    def _minmax(self):
        """
        :return: the min or max function, depending on the minmax parameter
        """
        if self.minmax == 'min':
            return np.min
        return np.max

    def _argminmax(self):
        """
        :return: the argmin or argmax function, depending on the minmax parameter
        """
        if self.minmax == 'min':
            return np.argmin
        return np.argmax

    def solve(self, problem_dict, verbose=False):
        """
        Solve the problem

        :param problem_dict: dictionary containing the problem definition
        :param verbose: if True, prints the best score at each epoch

        :return: None
        """
        if verbose:
            self.verbose = True
        self._before_initialization()
        self.initialize(problem_dict)
        for current_epoch in range(self.epoch):
            start = time.time()
            self.evolve(current_epoch)
            end = time.time()
            if self.verbose:
                print(f'Epoch: {current_epoch}, Best Score: {self.get_best_score()}')
            self.history.update_history(current_epoch, end - start)
            if self.history.is_early_stopping(current_epoch, self.early_stopping):
                if self.verbose:
                    print(f'Early stopping at epoch {current_epoch}')
                break
        return self.get_best_solution(), self.get_best_score()

    def get_history(self):
        """
        Get the history of the optimizer
        """
        return self.history.get_history()

    def visualize_history_fitness(self):
        """
        Visualize the fitness history
        """
        self.history.visualize_fitness()

    def visualize_history_time(self):
        """
        Visualize the time history
        """
        self.history.visualize_time()



from warnings import warn

from hyponic.optimizers.swarm_based.PSO import IWPSO
from hyponic.metrics.decorators import METRICS_DICT
from hyponic.utils.problem_identifier import ProblemIdentifier, ProblemType

from typing import Callable

from hyponic.space import Space
from functools import partial


class HypONIC:
    """
    HypONIC (Hyperparameter Optimization using Nature-Inspired Computing)

    Main class for hyperparameter optimization.
    """

    def __init__(self, model, X, y, metric: Callable | str | None = None, optimizer=None, **kwargs):
        self.model = model
        self.X = X
        self.y = y

        if isinstance(metric, str):
            # Try to get metric from the METRICS_DICT
            self.metric = METRICS_DICT.get(metric, None)

            if self.metric is None:
                raise Exception(f"Metric {metric} is not found.")
        elif isinstance(metric, Callable):
            self.metric = metric
        elif metric is None:
            # If metric is None, then try to get metric from the problem type
            problem_type = ProblemIdentifier(self.y).get_problem_type()

            match problem_type:
                case ProblemType.REGRESSION:
                    self.metric = METRICS_DICT["mse"]
                case ProblemType.BINARY_CLASSIFICATION:
                    self.metric = METRICS_DICT["binary_crossentropy"]
                case ProblemType.MULTICLASS_CLASSIFICATION:
                    self.metric = METRICS_DICT["log_loss"]
        else:
            raise Exception(f"Metric {metric} is not found.")

        try:
            self.minmax = self.metric.__getattribute__("minmax")
        except AttributeError:
            # If a metric does not have minmax attribute,
            # then it is assumed to be a custom metric and will be minimized by default
            warn(f"Metric {metric.__name__} does not have minmax attribute. Minimize by default.")
            self.minmax = "min"

        if kwargs is None:  # Default values for optimizer
            kwargs = {"epoch": 10, "pop_size": 10}

        if optimizer is None:
            self.optimizer = IWPSO(**kwargs)
        else:
            self.optimizer = optimizer(**kwargs)

        self.hyperparams_optimized = None
        self.metric_optimized = None

    @staticmethod
    def warn_not_optimized(func):
        """
        Decorator that warns if a method is called before optimization.
        """

        def wrapper(*args, **kwargs):
            if args[0].hyperparams_optimized is None:
                raise Exception("Model is not optimized yet. Please call optimize method first")
            return func(*args, **kwargs)

        return wrapper

    def _fitness_wrapper(self, dimensions_names: list, mapping_funcs: dict, values: list) -> float:
        # Map back to original space
        hyperparams = {
            dim: mapping_funcs[dim](val) for dim, val in zip(dimensions_names, values)
        }

        self.model.set_params(**hyperparams)
        self.model.fit(self.X, self.y)

        y_pred = self.model.predict(self.X)
        return self.metric(self.y, y_pred)  # TODO: maybe cross-validation could be used instead

    def optimize(self, hyperparams: dict, verbose=False) -> (dict, float):
        # Create a space for hyperparameters
        hyperspace = Space(hyperparams)
        if verbose:
            print("Successfully created a space for hyperparameters optimization")
            print(f"Using {self.optimizer.__class__.__name__} optimizer")
            print(f"Metric {self.metric.__name__} is subject to {self.minmax}imization")
            print(hyperspace)

        # Map hyperparameters to continuous space
        mappings_with_bounds = hyperspace.get_continuous_mappings(origins=0)  # Make that all dimensions start from 0

        # Split mappings and bounds
        mapping_funcs = {}

        low_bounds = []
        highs_bounds = []

        for name in hyperspace.dimensions_names:
            mapping, (low, high) = mappings_with_bounds[name]

            mapping_funcs[name] = mapping
            low_bounds.append(low)
            highs_bounds.append(high)

        paramspace = {
            "fit_func": partial(self._fitness_wrapper, hyperspace.dimensions_names, mapping_funcs),
            "lb": low_bounds,
            "ub": highs_bounds,
            "minmax": self.minmax
        }

        hyperparams_optimized, metric_optimized = self.optimizer.solve(paramspace, verbose=verbose)

        # Map back to the original space
        hyperparams_optimized = {
            dim: mapping_funcs[dim](val) for dim, val in zip(hyperspace.dimensions_names, hyperparams_optimized)
        }

        self.hyperparams_optimized = hyperparams_optimized
        self.metric_optimized = metric_optimized

        return hyperparams_optimized, metric_optimized

    @warn_not_optimized
    def get_optimized_model(self):
        self.model.set_params(**self.hyperparams_optimized)
        self.model.fit(self.X, self.y)
        return self.model

    @warn_not_optimized
    def get_optimized_parameters(self) -> dict:
        return self.hyperparams_optimized

    @warn_not_optimized
    def get_optimized_metric(self) -> float:
        return self.metric_optimized

    @warn_not_optimized
    def visualize_history_fitness(self):
        self.optimizer.visualize_history_fitness()

    @warn_not_optimized
    def visualize_history_time(self):
        self.optimizer.visualize_history_time()



"""
This module contains metrics for evaluating regression models.
"""
from hyponic.metrics.decorators import add_metric_aliases, add_metric_info, maximize_metric, minimize_metric

import numpy as np
import numexpr as ne


@add_metric_aliases("mean_absolute_error")
@add_metric_info("Mean Absolute Error (MAE) is the average of the absolute errors.")
@minimize_metric
def mae(y_true: np.array, y_pred: np.array) -> np.ndarray:
    ae = ne.evaluate("sum(abs(y_true - y_pred))")  # for technical reasons we have to separate summation and division
    return ne.evaluate("ae / length", local_dict={"ae": ae, "length": len(y_true)})


@add_metric_aliases("mean_squared_error")
@add_metric_info("Mean Squared Error (MSE) is the average of the squares of the errors.")
@minimize_metric
def mse(y_true: np.array, y_pred: np.array) -> np.ndarray:
    se = ne.evaluate("sum((y_true - y_pred) ** 2)")
    return ne.evaluate("se / length", local_dict={"se": se, "length": len(y_true)})


@add_metric_aliases("root_mean_squared_error")
@add_metric_info("Root Mean Squared Error (RMSE) is the square root of the average of the squared errors.")
@minimize_metric
def rmse(y_true: np.array, y_pred: np.array) -> np.ndarray:
    return np.sqrt(mse(y_true, y_pred))


@add_metric_aliases("root_mean_squared_log_error")
@add_metric_info("Root Mean Squared Log Error (RMSLE) is the log of the square root of"
                 "the average of the squared errors.")
@minimize_metric
def rmsle(y_true: np.array, y_pred: np.array) -> np.ndarray:
    return np.log(rmse(y_true, y_pred))


@maximize_metric
@add_metric_info("R2 score is the proportion of the variance in the dependent variable that is predictable")
def r2(y_true: np.array, y_pred: np.array) -> np.ndarray:
    ss_res = ne.evaluate("sum((y_true - y_pred) ** 2)")
    sum_y_true = ne.evaluate("sum(y_true)")
    mean = ne.evaluate("sum_y_true / length", local_dict={"sum_y_true": sum_y_true, "length": len(y_true)})
    ss_tot = ne.evaluate("sum((y_true - mean) ** 2)")
    return ne.evaluate("1 - ss_res / ss_tot")


@maximize_metric
@add_metric_info("Adjusted R2 score is the proportion of the variance in the dependent variable that is predictable")
def adjusted_r2(y_true: np.array, y_pred: np.array) -> np.ndarray:
    if len(y_true.shape) == 1:
        return r2(y_true, y_pred)
    return 1 - (1 - r2(y_true, y_pred)) * (len(y_true) - 1) / (len(y_true) - len(y_true[0]) - 1)


@minimize_metric
@add_metric_info("Huber loss is a loss function used in robust regression, that is less sensitive to outliers")
def huber_loss(y_true: np.array, y_pred: np.array, delta: float = 1.0) -> np.ndarray:
    error = y_true - y_pred
    return ne.evaluate("sum(where(abs(error) <= delta, 0.5 * error ** 2, delta * (abs(error) - 0.5 * delta)))")



__version__ = '0.1.1'

from .hyponic import HypONIC

from .metrics.decorators import add_metric_to_dict, add_metric_info,\
    add_metric_aliases, minimize_metric, maximize_metric

from .metrics.classification import accuracy, precision, recall, f1_score,\
    fbeta, confusion_matrix, binary_crossentropy, categorical_crossentropy, log_loss

from .metrics.regression import mae, mse, rmse, rmsle, r2, adjusted_r2, huber_loss

from .optimizers.genetic_based.GA import GA
from .optimizers.physics_based.SA import SA
from .optimizers.swarm_based.ABC import ABC
from .optimizers.swarm_based.ACO import ACO
from .optimizers.swarm_based.PSO import PSO, IWPSO
from .optimizers.swarm_based.GWO import GWO
from .optimizers.swarm_based.CS import CS
from .optimizers.base_optimizer import BaseOptimizer

from .space import Space, Dimension, Continuous, Discrete

from .utils.history import History
from .utils.problem_identifier import ProblemIdentifier, ProblemType



from hyponic.optimizers.base_optimizer import BaseOptimizer

import numpy as np
import numexpr as ne


class PSO(BaseOptimizer):
    """
    Particle Swarm Optimization (PSO) algorithm

    Hyperparameters:
        + a1(float), default=0.5: acceleration parameter
        + a2(float), default=0.5: acceleration parameter

    Example
    ~~~~~~~
    >>> from hyponic.optimizers.swarm_based.PSO import PSO
    >>> import numpy as np
    >>>
    >>> def sphere(x):
    >>>     return np.sum(x ** 2)
    >>>
    >>> problem_dict = {
    >>>     'fit_func': sphere,
    >>>     'lb': [-5.12, -5, -14, -6, -0.9],
    >>>     'ub': [5.12, 5, 14, 6, 0.9],
    >>>     'minmax': 'min'
    >>> }
    >>>
    >>> a1, a2 = 0.8, 0.4
    >>> pso = PSO(epoch=40, population_size=100, verbose=True, early_stopping=4, a1=a1, a2=a2)
    >>> pso.solve(problem_dict)
    >>> print(pso.get_best_score())
    >>> print(pso.get_best_solution())
    """
    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None, a1: float = 0.5, a2: float = 0.5,
                 verbose: bool = False, mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None,
                 **kwargs):
        """
        :param epoch: number of iterations
        :param population_size: number of individuals in the population
        :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
        :param verbose: whether to print the progress, default is False
        :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
        :param n_workers: number of workers to use in multithreading mode
        :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
        stopping is not used
        :param a1, a2: acceleration parameter
        """
        super().__init__(**kwargs)
        self.epoch = epoch
        self.population_size = population_size
        self.minmax = minmax
        self.verbose = verbose
        self.mode = mode
        self.n_workers = n_workers
        self.early_stopping = early_stopping

        self.a1 = a1
        self.a2 = a2
        self.velocities = None
        self.p_best = None
        self.p_best_coords = None
        self.g_best = None
        self.g_best_coords = None

    def _before_initialization(self):
        super()._before_initialization()
        if isinstance(self.a1, float) is False and isinstance(self.a1, int) is False:
            raise ValueError("a1 should be a float or an integer")
        if isinstance(self.a2, float) is False and isinstance(self.a2, int) is False:
            raise ValueError("a2 should be a float or an integer")

    def initialize(self, problem_dict):
        # TODO: if lb and ub are not provided, use the default values
        super().initialize(problem_dict)
        self.g_best = np.inf if self.minmax == "min" else -np.inf
        max_velocity = ne.evaluate("ub - lb", local_dict={'ub': self.ub, 'lb': self.lb})

        self.velocities = np.random.uniform(-max_velocity, max_velocity, size=(self.population_size, self.dimensions))
        self.p_best_coords = self.coords
        self.p_best = np.array([self.function(self.coords[i]) for i in range(self.population_size)])
        self._update_global_best()

    def evolve(self, epoch):
        self.velocities = self._update_velocity()
        self.coords = ne.evaluate("coords + velocities",
                                  local_dict={'coords': self.coords, 'velocities': self.velocities})

        # TODO: if lb or ub is provided, clip the coordinates
        self.coords = np.clip(self.coords, self.lb, self.ub)
        fitness = np.array([self.function(self.coords[i]) for i in range(self.population_size)])
        condition = all(self._minmax()(np.concatenate([self.p_best, fitness])) != self.p_best)

        self.p_best_coords = np.where(condition, self.coords, self.p_best_coords)

        self.p_best = ne.evaluate("where(condition, fitness, p_best)", local_dict={'condition': condition,
                                                                                   'fitness': fitness,
                                                                                   'p_best': self.p_best})
        self._update_global_best()

    def _update_velocity(self):
        r1 = np.random.random()
        r2 = np.random.random()
        expr = "velocities + a1 * r1 * (p_best_coords - coords) + a2 * r2 * (g_best_coords - coords)"
        return ne.evaluate(expr,
                           local_dict={'velocities': self.velocities, 'a1': self.a1, 'a2': self.a2, 'r1': r1, 'r2': r2,
                                       'p_best_coords': self.p_best_coords, 'coords': self.coords,
                                       'g_best_coords': self.g_best_coords})

    def _update_global_best(self):
        if self._minmax()(np.concatenate([self.p_best, [self.g_best]])) != self.g_best:
            self.g_best = self._minmax()(self.p_best)
            self.g_best_coords = self.p_best_coords[self._argminmax()(self.p_best)]

    def get_best_score(self):
        return self.g_best

    def get_best_solution(self):
        return self.g_best_coords

    def get_current_best_score(self):
        return self.p_best

    def get_current_best_solution(self):
        return self.p_best_coords


class IWPSO(PSO):
    """
    Inertia Weight Particle Swarm Optimization

    Hyperparameters:
        + a1(float), default=0.5: acceleration parameter
        + a2(float), default=0.5: acceleration parameter
        + w(float), default=0.5: inertia weight

    Example
    ~~~~~~~
    >>> from hyponic.optimizers.swarm_based.PSO import IWPSO
    >>> import numpy as np
    >>>
    >>> def sphere(x):
    >>>     return np.sum(x ** 2)
    >>>
    >>> problem_dict = {
    >>>     'fit_func': sphere,
    >>>     'lb': [-5.12, -5, -14, -6, -0.9],
    >>>     'ub': [5.12, 5, 14, 6, 0.9],
    >>>     'minmax': 'min'
    >>> }
    >>>
    >>> a1, a2 = 0.8, 0.4
    >>> w = 0.3
    >>> iwpso = IWPSO(epoch=40, population_size=100, verbose=True, early_stopping=4, a1=a1, a2=a2, w=w)
    >>> iwpso.solve(problem_dict)
    >>> print(iwpso.get_best_score())
    >>> print(iwpso.get_best_solution())
    """

    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None, a1: float = 0.5, a2: float = 0.5,
                 w: float = 0.8,
                 verbose: bool = False,
                 mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None, **kwargs):
        """
        :param epoch: number of iterations
        :param population_size: number of individuals in the population
        :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
        :param verbose: whether to print the progress, default is False
        :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
        :param n_workers: number of workers to use in multithreading mode
        :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
        stopping is not used
        :param a1, a2: acceleration parameter
        :param w: inertia
        """
        super().__init__(epoch, population_size, minmax, a1, a2, verbose, mode, n_workers, early_stopping, **kwargs)
        self.w = w

    def _before_initialization(self):
        super()._before_initialization()
        if isinstance(self.w, float) is False and isinstance(self.w, int) is False:
            raise ValueError("w should be a float or an integer")

    def _update_velocity(self):
        r1 = np.random.random()
        r2 = np.random.random()
        expr = "w * velocities + a1 * r1 * (p_best_coords - coords) + a2 * r2 * (g_best_coords - coords)"
        return ne.evaluate(expr, local_dict={'w': self.w, 'velocities': self.velocities, 'a1': self.a1, 'a2': self.a2,
                                             'r1': r1, 'r2': r2, 'p_best_coords': self.p_best_coords,
                                             'coords': self.coords, 'g_best_coords': self.g_best_coords})



import numpy as np
import matplotlib.pyplot as plt


class History:
    """
    This class is used to track the history of the optimizer. It tracks the global best, current best, and the time of
    the epoch.
    Also, it provides methods to visualize the history of the optimizer.

    """

    def __init__(self, **kwargs):
        self.optimizer = kwargs['optimizer']
        self.epoch = kwargs['epoch']
        self.population_size = kwargs['population_size']

        self.global_best_list = np.zeros((self.epoch, self.optimizer.dimensions))  # coordinates of the global best
        self.global_best_fitness_list = np.zeros(self.epoch)  # fitness of the global best
        self.current_best_list = np.zeros((self.epoch, self.optimizer.dimensions))  # coordinates of the current best
        self.current_best_fitness_list = np.zeros(self.epoch)  # fitness of the current best
        self.epoch_time_list = np.zeros(self.epoch)  # time of the current epoch

    def update_history(self, epoch, epoch_time):
        """
        This method updates the history of the optimizer
        """
        self.global_best_list[epoch] = self.optimizer.get_best_solution()
        self.global_best_fitness_list[epoch] = self.optimizer.get_best_score()
        self.current_best_list[epoch] = self.optimizer.get_best_solution()
        self.current_best_fitness_list[epoch] = self.optimizer.get_best_score()
        self.epoch_time_list[epoch] = epoch_time

    def get_history(self) -> dict:
        """
        This method returns the history of the optimizer
        """
        return {
            'global_best_list': self.global_best_list,
            'global_best_fitness_list': self.global_best_fitness_list,
            'current_best_list': self.current_best_list,
            'current_best_fitness_list': self.current_best_fitness_list,
            'epoch_time_list': self.epoch_time_list
        }

    def get_global_best(self) -> tuple:
        """
        This method returns the global best
        """
        return self.global_best_list[-1], self.global_best_fitness_list[-1]

    def get_current_best(self) -> tuple:
        """
        This method returns the current best
        """
        return self.current_best_list[-1], self.current_best_fitness_list[-1]

    @staticmethod
    def visualize(func):
        """
        Decorator to visualize the history of the optimizer

        :param func: function to be decorated
        """

        def wrapper(*args, **kwargs):
            plt.figure(figsize=(12, 8))
            func(*args, **kwargs)
            plt.legend()
            plt.show()

        return wrapper

    @visualize
    def visualize_fitness(self):
        """
        This method visualizes the history of the optimizer
        """
        plt.plot(self.global_best_fitness_list, label='Global Best')
        plt.plot(self.current_best_fitness_list, label='Current Best')
        plt.xlabel('Epochs')
        plt.ylabel('Fitness')
        plt.title('Fitness vs Epochs')

    @visualize
    def visualize_time(self):
        """
        This method visualizes the history of the optimizer
        """
        plt.plot(self.epoch_time_list, label='Epoch Time')
        plt.xlabel('Epochs')
        plt.ylabel('Time')
        plt.title('Time vs Epochs')

    def is_early_stopping(self, current_epoch, early_stopping: int):
        """
        This method checks if the early stopping condition is met

        :param current_epoch: current epoch
        :param early_stopping: number of epochs to wait before stopping the optimization

        :return: boolean
        """
        if early_stopping is not None and current_epoch >= early_stopping:
            last_scores = self.global_best_fitness_list[current_epoch - early_stopping + 1:current_epoch + 1]
            if len(last_scores) == early_stopping and np.all(last_scores == last_scores[0]):
                return True
        return False



from hyponic.optimizers.base_optimizer import BaseOptimizer

import numpy as np
import numexpr as ne


class CS(BaseOptimizer):
    """
    Cuckoo Search (CS) algorithm

    Hyperparameters:
        + pa(float), default=0.25: probability of cuckoo's egg to be discovered
        + alpha(float), default=0.5: step size
        + k(float), default=1: Levy multiplication coefficient

    Example
    ~~~~~~~
    >>> from hyponic.optimizers.swarm_based.CS import CS
    >>> import numpy as np
    >>>
    >>> def sphere(x):
    >>>     return np.sum(x ** 2)
    >>>
    >>> problem_dict = {
    >>>     'fit_func': sphere,
    >>>     'lb': [-5.12, -5, -14, -6, -0.9],
    >>>     'ub': [5.12, 5, 14, 6, 0.9],
    >>>     'minmax': 'min'
    >>> }
    >>>
    >>> cs = CS(epoch=40, population_size=100, verbose=True, early_stopping=4)
    >>> cs.solve(problem_dict)
    >>> print(cs.get_best_score())
    >>> print(cs.get_best_solution())
    """

    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None, verbose: bool = False,
                 pa: float = 0.25, alpha: float = 0.5, k: float = 1,
                 mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None, **kwargs):
        """
        :param epoch: number of iterations
        :param population_size: number of individuals in the population
        :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
        :param verbose: whether to print the progress, default is False
        :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
        :param n_workers: number of workers to use in multithreading mode
        :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
        stopping is not used
        :param pa: probability of cuckoo's egg to be discovered
        :param alpha: step size
        :param k: Levy multiplication coefficient
        """
        super().__init__(**kwargs)
        self.epoch = epoch
        self.population_size = population_size
        self.minmax = minmax
        self.verbose = verbose
        self.mode = mode
        self.n_workers = n_workers
        self.early_stopping = early_stopping

        self.pa = pa
        self.alpha = alpha
        self.k = k

        self.nests = None
        self.nests_fitness = None
        self.cuckoo_coords = None

    def _before_initialization(self):
        super()._before_initialization()
        if isinstance(self.pa, float) is False and isinstance(self.pa, int) is False:
            raise TypeError('pa should be a float or an integer')
        if isinstance(self.alpha, float) is False and isinstance(self.alpha, int) is False:
            raise TypeError('alpha should be a float or an integer')
        if isinstance(self.k, float) is False and isinstance(self.k, int) is False:
            raise TypeError('k should be a float or an integer')

    def initialize(self, problem_dict):
        super().initialize(problem_dict)

        self.nests = self.coords
        self.nests_fitness = np.array([self.function(self.nests[i]) for i in range(self.population_size)])
        self.cuckoo_coords = np.random.uniform(self.lb, self.ub, self.dimensions)

    def _levy_flight(self, x):
        u = np.random.normal(0, 1, size=self.dimensions)
        v = np.random.normal(0, 1, size=self.dimensions)
        best_coords = self.nests[self._argminmax()(self.nests_fitness)]
        return ne.evaluate('x + k * u / (abs(v) ** (1 / 1.5)) * (best_coords - x)', local_dict={
            'x': x, 'k': self.k, 'u': u, 'v': v, 'best_coords': best_coords
        })

    def evolve(self, current_epoch):
        x_new = self._levy_flight(self.cuckoo_coords)
        self.cuckoo_coords = np.clip(x_new, self.lb, self.ub)

        next_nest = np.random.randint(0, self.population_size)
        new_fitness = self.function(self.cuckoo_coords)
        if new_fitness < self.nests_fitness[next_nest]:
            self.nests[next_nest] = self.cuckoo_coords
            self.nests_fitness[next_nest] = new_fitness
        number_of_discovered_eggs = int(self.pa * self.population_size)
        worst_nests = np.argsort(self.nests_fitness)[-number_of_discovered_eggs:]
        self.nests[worst_nests] = np.random.uniform(self.lb, self.ub, (number_of_discovered_eggs, self.dimensions))
        self.nests_fitness[worst_nests] = np.array([self.function(self.nests[i]) for i in worst_nests])

    def get_best_score(self):
        return self._minmax()(self.nests_fitness)

    def get_best_solution(self):
        return self.nests[self._argminmax()(self.nests_fitness)]

    def get_current_best_score(self):
        return self.get_best_score()

    def get_current_best_solution(self):
        return self.get_best_solution()



from hyponic.optimizers.base_optimizer import BaseOptimizer

import numpy as np
import numexpr as ne


class SA(BaseOptimizer):
    """
    Simulated Annealing

    Example
    ~~~~~~~
    >>> from hyponic.optimizers.physics_based.SA import SA
    >>> import numpy as np
    >>>
    >>> def sphere(x):
    >>>     return np.sum(x ** 2)
    >>>
    >>> problem_dict = {
    >>>     'fit_func': sphere,
    >>>     'lb': [-5.12, -5, -14, -6, -0.9],
    >>>     'ub': [5.12, 5, 14, 6, 0.9],
    >>>     'minmax': 'min'
    >>> }
    >>>
    >>> sa = SA(epoch=40, population_size=100, verbose=True, early_stopping=4)
    >>> sa.solve(problem_dict)
    >>> print(sa.get_best_score())
    >>> print(sa.get_best_solution())
    """

    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None,
                 verbose: bool = False, mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None,
                 **kwargs):
        """
        :param epoch: number of iterations
        :param population_size: number of individuals in the population
        :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
        :param verbose: whether to print the progress, default is False
        :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
        :param n_workers: number of workers to use in multithreading mode
        :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
        stopping is not used
        """
        super().__init__(**kwargs)
        self.epoch = epoch
        self.population_size = population_size
        self.minmax = minmax
        self.verbose = verbose
        self.mode = mode
        self.n_workers = n_workers
        self.early_stopping = early_stopping

        self.currents = None
        self.best = None
        self.best_score = None

    def initialize(self, problem_dict):
        super().initialize(problem_dict)

        self.currents = np.random.uniform(self.lb, self.ub, (self.population_size, self.dimensions))

        self.best = self.currents[0]
        self.best_score = self.function(self.best)

    def evolve(self, current_epoch):
        progress = current_epoch / self.epoch
        t = max(0.01, min(1, 1 - progress))

        amplitudes = ne.evaluate("(_max - _min) * progress * 0.1",
                                 local_dict={'_max': np.max(self.intervals), '_min': np.min(self.intervals),
                                             'progress': progress})
        deltas = np.random.uniform(-amplitudes / 2, amplitudes / 2, (self.population_size, self.dimensions))

        candidates = ne.evaluate("currents + deltas", local_dict={'currents': self.currents, 'deltas': deltas})

        for idx, candidate in enumerate(candidates):
            candidate = np.clip(candidate, self.lb, self.ub)
            candidate_score = self.function(candidate)

            if candidate_score < self.best_score:  # TODO: check if the problem is minimization or maximization
                self.best = candidate
                self.best_score = candidate_score
                self.currents[idx] = candidate
            else:
                score_abs_diff = ne.evaluate("abs(candidate_score - best_score)",
                                             local_dict={'candidate_score': candidate_score,
                                                         'best_score': self.best_score})

                if np.random.uniform() < np.exp(-score_abs_diff / t):
                    self.currents[idx] = candidate

    def get_best_score(self):
        return self.best_score

    def get_best_solution(self):
        return self.best

    def get_current_best_score(self):
        return self.function(self.currents[0])

    def get_current_best_solution(self):
        return self.currents[0]



import setuptools


def readme():
    with open('README.md') as f:
        return f.read()


setuptools.setup(
    name='hyponic',
    version='0.1.1',
    author='Vladislav Kulikov, Daniel Satarov, Ivan Chernakov',
    author_email='v.kulikov@innopolis.university, d.satarov@innopolis.university, i.chernakov@innopolis.university',
    description='Hyperparameter Optimization with Nature-Inspired Computing',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/slewie/HypONIC',
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.23.5',
        'numexpr>=2.8.4',
        'numba>=0.57.0',
        'matplotlib>=3.6.3',
    ],
    classifiers=[
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)



class Symbol:
    def __init__(self, symtable, name, *args, **kwargs):
        self._symtable = symtable
        self._name = name
        self._args = args
        self._kwargs = kwargs  # TODO: Implement kwargs

    def __call__(self, *args):
        return Symbol(self._symtable, self._name, *args)

    def __repr__(self):
        return f"{self._name}({', '.join(repr(arg) for arg in self._args)})"

    def eval(self):
        args = [arg.eval() if isinstance(arg, Symbol) else arg for arg in self._args]

        # If all arguments are literals, evaluate the function
        # Otherwise, evaluate partially
        if all(not isinstance(arg, Symbol) for arg in args):
            if self._name in self._symtable.get_symbol_list():
                return self._symtable.get_symbol(self._name)(*args)
            else:
                return self
        else:
            # Partially evaluate
            self._args = args
            return self


class SymbolTable:
    def __init__(self):
        self._symbols = {
            "int": int,
            "float": float,
            "add": lambda x, y: x + y,
        }

    def get_symbol(self, name):
        return self._symbols[name]

    def get_symbol_list(self):
        return self._symbols.keys()

    def add_symbol(self, name, func):
        self._symbols[name] = func

    def __getattr__(self, name):
        return Symbol(self, name)


# scope = SymbolTable()
# expr = scope.add(scope.int(1), scope.int(2.5))
# expr.eval()



import numpy as np
from enum import Enum


class ProblemType(Enum):
    """
    This class is used to identify the problem type using dataset. It can be used for regression and classification only
    """
    REGRESSION = 1
    BINARY_CLASSIFICATION = 2
    MULTICLASS_CLASSIFICATION = 3


class ProblemIdentifier:
    """
    This class is used to identify the problem type using dataset. It can be used for regression and classification only
    """

    def __init__(self, y):
        self.y = y

    def get_problem_type(self):
        """
        This method identifies the problem type using number of classes and data type of y.
        If number of classes is 2 and data type of y is int64, then it is a binary classification problem
        If number of classes is more than 2 and data type of y is int64, then it is a multiclass classification problem
        If data type of y is not int64, then it is a regression problem
        """
        target_type = type(self.y[0])
        number_of_classes = len(np.unique(self.y))
        if target_type in [np.int64, np.int32, np.int16, np.int8]:
            if number_of_classes == 2:
                return ProblemType.BINARY_CLASSIFICATION
            else:
                return ProblemType.MULTICLASS_CLASSIFICATION
        else:
            return ProblemType.REGRESSION



from hyponic.optimizers.base_optimizer import BaseOptimizer

import numpy as np


class ACO(BaseOptimizer):
    """
    Ant Colony Optimization (ACO)

    Hyperparameters:
        + alpha(float), default=1: the importance of pheromone
        + beta(float), default=1: the importance of heuristic information
        + rho(float), default=0.5: the pheromone evaporation rate
        + q(float), default=1: the pheromone intensity

    Example
    ~~~~~~~
    >>> from hyponic.optimizers.swarm_based.ACO import ACO
    >>> import numpy as np
    >>>
    >>> def sphere(x):
    >>>     return np.sum(x ** 2)
    >>>
    >>> problem_dict = {
    >>>     'fit_func': sphere,
    >>>     'lb': [-5.12, -5, -14, -6, -0.9],
    >>>     'ub': [5.12, 5, 14, 6, 0.9],
    >>>     'minmax': 'min'
    >>> }
    >>>
    >>> alpha = 1
    >>> beta = 1
    >>> rho = 0.5
    >>> q = 1
    >>>
    >>> aco = ACO(epoch=40, population_size=100, verbose=True, early_stopping=4, alpha=alpha, beta=beta, rho=rho, q=q)
    >>> aco.solve(problem_dict)
    >>> print(aco.get_best_score())
    >>> print(aco.get_best_solution())
    """

    def __init__(self, epoch: int = 10, population_size: int = 10, minmax: str = None, alpha: float = 1,
                 beta: float = 1, rho: float = 0.5, q: float = 1,
                 verbose: bool = False, mode: str = 'single', n_workers: int = 4, early_stopping: int | None = None,
                 **kwargs):
        """
        :param epoch: number of iterations
        :param population_size: number of individuals in the population
        :param minmax: 'min' or 'max', depending on whether the problem is a minimization or maximization problem
        :param verbose: whether to print the progress, default is False
        :param mode: 'single' or 'multithread', depending on whether to use multithreading or not
        :param n_workers: number of workers to use in multithreading mode
        :param early_stopping: number of epochs to wait before stopping the optimization process. If None, then early
        stopping is not used
        :param alpha: the importance of pheromone
        :param beta: the importance of heuristic information
        :param rho: the pheromone evaporation rate
        :param q: the pheromone intensity
        """
        super().__init__(**kwargs)
        self.epoch = epoch
        self.population_size = population_size
        self.minmax = minmax
        self.verbose = verbose
        self.mode = mode
        self.n_workers = n_workers
        self.early_stopping = early_stopping

        self.alpha = alpha
        self.beta = beta
        self.rho = rho
        self.q = q

        self.pheromone = None
        self.population = None
        self.scores = None
        self.best_score = None
        self.best_solution = None

    def _before_initialization(self):
        super()._before_initialization()
        if self.alpha < 0:
            raise ValueError("alpha should be a positive float")
        if self.beta < 0:
            raise ValueError("beta should be a positive float")
        if self.rho < 0 or self.rho > 1:
            raise ValueError("rho should be a float between 0 and 1")
        if self.q < 0:
            raise ValueError("q should be a positive float")

    def initialize(self, problem_dict):
        super().initialize(problem_dict)

        self.pheromone = np.ones((self.population_size, self.dimensions))
        self.best_score = np.inf if self.minmax == 'min' else -np.inf

        self.population = np.random.uniform(low=self.lb, high=self.ub, size=(self.population_size, self.dimensions))
        self.scores = np.zeros(self.population_size)
        for i in range(self.population_size):
            self.scores[i] = self.function(self.population[i])
            if self._minmax()(self.scores[i]) < self._minmax()(self.best_score):
                self.best_score = self.scores[i]
                self.best_solution = self.population[i]

    def evolve(self, epoch):
        new_population = np.zeros((self.population_size, self.dimensions))
        new_scores = np.zeros(self.population_size)

        for i in range(self.population_size):
            ant_position = np.random.uniform(low=self.lb, high=self.ub, size=self.dimensions)

            for j in range(self.dimensions):
                pheromone_weights = self.pheromone[:, j] ** self.alpha
                heuristic_weights = 1 / (np.abs(ant_position[j] - self.population[:, j]) + 1e-6) ** self.beta
                probs = pheromone_weights * heuristic_weights
                probs /= np.sum(probs)
                ant_position[j] = np.random.choice(self.population[:, j], p=probs)

            ant_score = self.function(ant_position)
            self._update_pheromone(ant_score)

            if self._minmax()(ant_score) < self._minmax()(self.best_score):
                self.best_solution = ant_position
                self.best_score = ant_score

            new_population[i] = ant_position
            new_scores[i] = ant_score

        self.population = new_population
        self.scores = new_scores

    def get_best_score(self):
        return self.best_score

    def get_best_solution(self):
        return self.best_solution

    def get_current_best_score(self):
        return self._minmax()(self.scores.min())

    def get_current_best_solution(self):
        return self.population[self.scores.argmin()]

    def _initialize_pheromone(self):
        self.pheromone = np.ones((self.population_size, self.dimensions))

    def _update_pheromone(self, ant_score):
        delta_pheromone = np.zeros((self.population_size, self.dimensions))
        delta_pheromone[:, :] = self.q / ant_score
        self.pheromone = (1 - self.rho) * self.pheromone + self.rho * delta_pheromone
